/*
 * OCSP responder
 * by Massimiliano Pala (madwolf@openca.org)
 * OpenCA project 2001
 *
 * Copyright (c) 2001 The OpenCA Project.  All rights reserved.
 *
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include "general.h"
#include "support.h"

#include "crl.h"
#include "ocsp_db.h"

#define _XOPEN_SOURCE 500

#include <unistd.h>

/* External imported variables */
extern OCSPD_CONFIG * ocspd_conf;
extern pthread_rwlock_t crl_lock;
extern pthread_cond_t crl_cond;

int
ocspd_load_ca_crl(CA_LIST_ENTRY *a)
{

	const ASN1_TIME *lastup;
	const ASN1_TIME *nextup;

	if (!a)
		return -1;

	if (ocspd_conf->debug)
		fprintf(stderr, 
			"INFO::ACQUIRING WRITE LOCK -- BEGIN CRL RELOAD");
	pthread_rwlock_wrlock(&crl_lock);
	if (ocspd_conf->debug)
		fprintf(stderr, "INFO::LOCK ACQUIRED (CRL RELOAD)");

	if (a->crl)
		X509_CRL_free(a->crl);

	a->crl = NULL;
	a->crl_list = NULL;

	if (a->crl_url == NULL) {
		syslog(LOG_ERR, "ERROR::Missing CRL URL for CA %s", a->ca_id);
		return -1;
	}

	/* We now re-load the CRL */
	if( (a->crl = ocspd_get_crl( a->crl_url )) == NULL ) {
		syslog( LOG_ERR, "ERROR::Can not reload CRL [ %s ] for CA "
			"[%s]", a->crl_url->addr, a->ca_id);
		return(-1);
	}

	if( ocspd_conf->verbose )
		syslog( LOG_INFO, "INFO::CRL successfully reloaded [ %s ]",
			a->ca_id );

	/* Let's get the CRLs entries, if any */
	if( ocspd_build_crl_entries_list ( a, a->crl ) == NULL ) { 
		if( ocspd_conf->verbose )
			syslog(LOG_INFO, "INFO::No Entries for CRL [ %s ]",
				a->ca_id );
	};

	if(ocspd_conf->verbose)
		syslog( LOG_INFO, 
			"INFO::CRL loaded successfully [ %s ]", a->ca_id );

	/* If previous values are there, then we clear them up */
	if ( a->lastUpdate ) ASN1_TIME_free(a->lastUpdate);
	if ( a->nextUpdate ) ASN1_TIME_free(a->nextUpdate);

	/* Get new values from the recently loaded CRL */
#if OPENSSL_VERSION_NUMBER < 0x10100005L
	lastup = X509_CRL_get_lastUpdate(a->crl);
	nextup = X509_CRL_get_nextUpdate(a->crl);
#else
	lastup = X509_CRL_get0_lastUpdate(a->crl);
	nextup = X509_CRL_get0_nextUpdate(a->crl);
#endif

	a->lastUpdate = (ASN1_TIME *) ASN1_STRING_dup((ASN1_STRING *) lastup);
	a->nextUpdate = (ASN1_TIME *) ASN1_STRING_dup((ASN1_STRING *) nextup);

	if (ocspd_conf->debug)
		fprintf(stderr, "INFO::RELEASING LOCK (CRL RELOAD)");
	pthread_rwlock_unlock(&crl_lock);
	if (ocspd_conf->debug)
		fprintf(stderr, "INFO::LOCK RELEASED --END--");

	/* Now check the CRL validity */
	a->crl_status = check_crl_validity( a );

	if( a->crl_status == CRL_OK ) {
		syslog(LOG_ERR, "INFO::CRL for CA [%s] (OK)",
			a->ca_id);
	}

	return(0);
}


int ocspd_reload_crls ( OCSPD_CONFIG *conf ) {

	int i;

	CA_LIST_ENTRY *a = NULL;

	syslog(LOG_INFO, "INFO::CRL Reload --START--");

	if( conf->verbose )
		syslog(LOG_INFO, "INFO::CRL Reload %ld CAs",
			conf->ca_list_len );

	for( i=0; i < conf->ca_list_len; i++ ) {
		a = conf->ca_list[i];

		syslog(LOG_INFO, "INFO::Reloading CRL for CA [%s]",
			a->ca_id );

		if( ocspd_load_ca_crl(a) < 0 ) {
			syslog( LOG_ERR, "ERROR::Reload CRL for CA [%s]",
					a->ca_id );
			continue;
		}
	}

	syslog(LOG_INFO, "INFO::CRL Reload --END--");

	return(1);
}

int check_crl ( X509_CRL *crl, STACK_OF(X509) *cacert ) {

	EVP_PKEY *pkey = NULL;
	int ret = -1;
	int i, final;

	if( ocspd_conf->verbose ) {
		syslog(LOG_INFO,"INFO::Checking CRL");
	};

	pthread_rwlock_rdlock( &crl_lock );
	if( !crl || !cacert ) {
		if( ocspd_conf->verbose ) {
			if(!crl)
				syslog(LOG_ERR, "CRL missing");
			if(!cacert)
				syslog(LOG_ERR, "CA cert missing");
		}
		pthread_rwlock_unlock( &crl_lock );
		return(-1);
	}

	if( sk_X509_num(cacert) < 1 ) {
		if(ocspd_conf->verbose)
			syslog(LOG_ERR, "No CA cert loaded (%d)", 
				sk_X509_num(cacert));
		pthread_rwlock_unlock( &crl_lock );
		return(-2);
	}

	final = -99;
	for( i = 0; i < sk_X509_num(cacert); i++ ) {
		X509 *x = NULL;

		if( (x = sk_X509_value(cacert,i)) == NULL ) {
			continue;
		}
		pkey = X509_get_pubkey( x );

		/* Gets the Public Key of the CA Certificate */
		if(pkey == NULL ) {
			syslog( LOG_ERR, "ERROR parsing Pub Key from CA Cert "
				" [%d]", i );
			pthread_rwlock_unlock( &crl_lock );
			return(-3);
		}

		/* Ever CLEAR the error queue!!!! */
		ERR_clear_error();

		/* Checks the CRL  - 0 if failure, > 0 if successful */
		ret = X509_CRL_verify(crl, pkey);

		/* Free allocked memory */
		// if( pkey ) EVP_PKEY_free (pkey);

		if( ret > final ) final = ret;

		if( ocspd_conf->verbose ) {
			if ( ret > 0 ) {
			   syslog(LOG_INFO, "CRL and CA cert [%d:%d] check ok",
				i, ret );
			} else {
			   syslog(LOG_INFO, "CRL and CA cert [%d:%d] not ok",
				i, ret );
			}
		}
	}
	pthread_rwlock_unlock( &crl_lock );

	if ( final > 0 ) {
		if(ocspd_conf->verbose)
			syslog(LOG_INFO, "CRL matching CA cert ok [ %d ]",ret);
	}

	return final;
}

int check_crl_validity ( CA_LIST_ENTRY *ca ) {
	int i;

	pthread_rwlock_rdlock( &crl_lock );

	if( (!ca) || (!ca->crl) || (!(ca->lastUpdate)) ) {
		syslog(LOG_ERR, "ERROR::CRL::MEMORY [%s] "
			"Verify error (memory alloc)", ca->ca_id );
		pthread_rwlock_unlock( &crl_lock );
		return(CRL_ERROR_LAST_UPDATE);
	}

        i=X509_cmp_time(ca->lastUpdate, NULL);
        if (i == 0) {
		syslog(LOG_ERR, "ERROR::CRL [%s] LAST UPDATE "
			"error (code %d)", ca->ca_id,
			CRL_ERROR_LAST_UPDATE );

		pthread_rwlock_unlock( &crl_lock );
		return(CRL_ERROR_LAST_UPDATE);
        }
        if (i > 0) {
		syslog(LOG_ERR, "WARING::CRL [%s] NOT YET "
			"valid (code %d)", ca->ca_id,
				CRL_NOT_YET_VALID);
		pthread_rwlock_unlock( &crl_lock );
		return(CRL_NOT_YET_VALID);
        }
                                                                                
        if(ca->nextUpdate) {
		i=X509_cmp_time(ca->nextUpdate, NULL);
                                                                                
                if (i == 0) {
			syslog(LOG_ERR, "ERROR::CRL [%s] NEXT UPDATE "
				"error (code %d)", ca->ca_id, 
					CRL_ERROR_NEXT_UPDATE );
			pthread_rwlock_unlock( &crl_lock );
			return(CRL_ERROR_NEXT_UPDATE);
		}
                                                                                
                if (i < 0) {
			syslog(LOG_WARNING, 
				"WARNING::CRL [%s] IS EXPIRED (code %d)",
					ca->ca_id, CRL_EXPIRED );
			pthread_rwlock_unlock( &crl_lock );
			return(CRL_EXPIRED);
                }
	} else {
		syslog( LOG_WARNING, "WARNING::CRL [%s] has no "
			"nextUpdate!", ca->ca_id );
	}
	pthread_rwlock_unlock( &crl_lock );

	if( ocspd_conf->verbose )
		syslog(LOG_INFO, "INFO::CRL::Verify %d [OK=%d]",
			i, CRL_OK );

	return (CRL_OK);
}

char * get_crl_status_info ( int status ) {

	switch( status ) {
		case CRL_OK:
			return("CRL is VALID");
			break;
			;;
		case CRL_ERROR_LAST_UPDATE:
			return("ERROR in LAST UPDATE field");
			break;
			;;
		case CRL_NOT_YET_VALID:
			return("WARNING, CRL is NOT YET valid");
			break;
			;;
		case CRL_ERROR_NEXT_UPDATE:
			return("ERROR in NEXT UPDATE field");
			break;
			;;
		case CRL_EXPIRED:
			return("CRL is EXPIRED");
			break;
			;;
	}
	return("CRL status is UNKNOWN");
}

void auto_crl_check ( int sig ) {

	CA_LIST_ENTRY *ca = NULL;
	int i, ret;

	if( ocspd_conf->verbose == 1 ) {
		syslog(LOG_INFO, "INFO::auto_crl_check() started");
	}

	if( ocspd_conf->crl_auto_reload ) {
		ocspd_conf->current_crl_reload += 
					ocspd_conf->alarm_decrement;

		if( ocspd_conf->current_crl_reload >=
					ocspd_conf->crl_auto_reload ) {

			ocspd_conf->current_crl_reload = 0;

			/* Here we de-allocate the CRL entries and
			   reload the CRL */
			if( ocspd_reload_crls( ocspd_conf ) == 0 ) {
				syslog( LOG_ERR, "Error reloading CRLs");
			} else {
				if( ocspd_conf->verbose )
					syslog( LOG_INFO, "CRLs reloaded.");
			}

			alarm( ocspd_conf->alarm_decrement );

			return;
		}
	}

	if( ocspd_conf->verbose == 1 ) {
		syslog(LOG_INFO, "INFO::auto_crl_check() continuing");
	}

	for( i=0; i < ocspd_conf->ca_list_len; i++ ) {

		if((ca = ocspd_conf->ca_list[i]) == NULL ) {
			continue;
		}

		if( ocspd_conf->verbose && ca->ca_id )
			syslog( LOG_INFO, "Auto CRL checking [%s]", ca->ca_id);


		ret = check_crl_validity ( ca );

		if( ca->crl_status != ret ) {
			if(ocspd_conf->verbose) 
				syslog( LOG_INFO,"Detected CRL status change");
			ca->crl_status = ret;

			ocspd_load_ca_crl (ca);

			continue;
		} else {
			if( ocspd_conf->verbose && ca->ca_id ) 
				syslog( LOG_INFO,"No CRL status change for [%s]",
					ca->ca_id);
		}
	}

	/*
	if( ocspd_conf->crl_check_validity ) {
		if( verbose )
			syslog(LOG_INFO, "Checking again CRL in %d secs",
				ocspd_conf->crl_check_validity );

		alarm( ocspd_conf->crl_check_validity );
	}
	*/

	if( ocspd_conf->verbose == 1 ) {
		syslog(LOG_INFO, "INFO::auto_crl_check() completed");
	}

	alarm( ocspd_conf->alarm_decrement );

	return;
}

