/* src/ocsp_response.c
 * ============================================================
 * OCSP Responder
 * (c) 2001-2006 by Massimiliano Pala and OpenCA Group
 *     OpenCA License software
 * ============================================================
 */
 
#include "general.h"

#include <openssl/err.h>
#include <openssl/ssl.h>

#include <openssl/txt_db.h>
#include <openssl/x509.h>

#include "support.h"
#include "ocsp_db.h"

#include "ocsp_response.h"
#include "configuration.h"
#include "crl.h"

pthread_mutex_t sign_mutex = PTHREAD_MUTEX_INITIALIZER;

OCSP_RESPONSE *make_ocsp_response( OCSP_REQUEST *req,
			OCSPD_CONFIG *conf, EVP_PKEY *pkey ) {

	OCSP_CERTID *cid = NULL;
	OCSP_BASICRESP *bs = NULL;
	OCSP_RESPONSE *resp = NULL;

	int i, id_count;

	ASN1_GENERALIZEDTIME *thisupd = NULL;
	ASN1_GENERALIZEDTIME *nextupd = NULL;

	id_count = OCSP_request_onereq_count(req);

	if (id_count <= 0) {
		syslog(LOG_ERR, "[Thread] ERROR::Request has no onereq "
				"(OpenSSL Error is -> %s)",
					ERR_error_string(ERR_get_error(),NULL));

		resp = OCSP_response_create(
				OCSP_RESPONSE_STATUS_MALFORMEDREQUEST, NULL);
		goto end;
	}

	bs = OCSP_BASICRESP_new();
	if (conf->set_nextUpdate)
		nextupd = X509_gmtime_adj(NULL, conf->nmin * 60 + 
				conf->ndays * 3600 * 24 );

	/* Examine each certificate id in the request */
	for (i = 0; i < id_count; i++) {
		OCSP_ONEREQ *one = NULL;
		ASN1_INTEGER *serial = NULL;
		CA_LIST_ENTRY *ca = NULL;
		X509_REVOKED *entry = NULL;

		one = OCSP_request_onereq_get0(req, i);
		cid = OCSP_onereq_get0_id(one);

		/* Get basic request info */
		OCSP_id_get0_info(NULL, NULL, NULL, &serial, cid);

		if( conf->verbose ) {
			syslog( LOG_INFO, "request for certificate serial %ld",
					ASN1_INTEGER_get( serial ) );
		}

		/* Is this request about our CA? */
		if ((ca = ocspd_CA_ENTRY_find( conf, cid )) == NULL) {
			if (conf->verbose) {
				syslog( LOG_INFO, 
				  "request for non recognized CA [serial %ld]",
					ASN1_INTEGER_get(serial) );
			}
			OCSP_basic_add1_status(bs, cid,
				V_OCSP_CERTSTATUS_UNKNOWN,0,NULL,
				X509_gmtime_adj(NULL, 0),nextupd);
			continue;
		}

		if( ca->compromised > 0 ) {
			OCSP_basic_add1_status( bs, cid,
				V_OCSP_CERTSTATUS_REVOKED, 0, NULL,
				X509_gmtime_adj(NULL, 0),nextupd);
			continue;
		}

		if( ca->crl_status != CRL_OK ) {
			resp = OCSP_response_create(
				OCSP_RESPONSE_STATUS_TRYLATER, NULL);
			syslog(LOG_ERR, "SENT TRYLATER (%s)", 
					get_crl_status_info (ca->crl_status));
			goto end;
		}

		/* Get the entry from the CRL data, if NULL then the
		   certificate is not revoked */
		entry = ocspd_X509_REVOKED_find( ca, serial );

		/* Sets thisUpdate field to the value of the loaded CRL */
		thisupd = ca->lastUpdate;

		if (entry) {
			OCSP_SINGLERESP *single = NULL;
			int reason = -1;
			ASN1_ENUMERATED *asn;
			const ASN1_TIME *rdate;
			void *ext;

			/* If reason is given, get it. */
			if ((asn = X509_REVOKED_get_ext_d2i(entry, 
					NID_crl_reason,NULL,NULL))) {
				reason = ASN1_ENUMERATED_get(asn);
				ASN1_ENUMERATED_free(asn);
			}

#if OPENSSL_VERSION_NUMBER < 0x10100005L
			rdate = entry->revocationDate;
#else
			rdate = X509_REVOKED_get0_revocationDate(entry);
#endif

			single = OCSP_basic_add1_status(bs, cid,
						V_OCSP_CERTSTATUS_REVOKED,
						reason,  (ASN1_TIME *) rdate,
						thisupd, nextupd);

			if (!single)
				syslog(LOG_ERR, "ERROR::Can not generate"
				       " basic response");

			/* Check and add the invalidity date */
			ext = X509_REVOKED_get_ext_d2i(entry, 
				NID_invalidity_date, NULL, NULL);
			if (ext)
				OCSP_SINGLERESP_add1_ext_i2d(single, 
					NID_invalidity_date, ext, 0, 0);

			if (conf->verbose)
				syslog(LOG_INFO, "Status for %ld is REVOKED",
						ASN1_INTEGER_get(serial));
		} else {
			if (conf->verbose)
				syslog(LOG_INFO, "status VALID for %ld",
				       ASN1_INTEGER_get(serial));
			
			OCSP_basic_add1_status(bs, cid,
				V_OCSP_CERTSTATUS_GOOD, 0, NULL, 
				thisupd, nextupd);
		}
	}

	OCSP_copy_nonce(bs, req);

	/* It seems that this function is not thread safe!!! */
	if ( pkey != NULL ) {
		if ( conf->ocspd_cert == 0 ) {
			syslog(LOG_ERR, "[%s:%d] ERROR::OCSP_CERT is NULL!",
				__FILE__, __LINE__ );
			if( resp ) OCSP_RESPONSE_free ( resp );
			resp = NULL;

			goto end;
		}

		if ( bs == NULL ) {
			syslog(LOG_ERR, "[%s:%d] ERROR::BS is NULL!",
				__FILE__, __LINE__ );
			if( resp ) OCSP_RESPONSE_free ( resp );
			resp = NULL;

			goto end;
		}

		// pthread_mutex_lock ( &sign_mutex );
		/* Sign the response only if we have a valid pkey pointer! */
		if(OCSP_basic_sign(bs, conf->ocspd_cert, pkey, 
			conf->digest, conf->other_certs, conf->flags) == 0 ) {
			syslog(LOG_ERR, "[Thread] ERROR::Response "
				"Signature Failed (OpenSSL Error is -> %s)",
				ERR_error_string(ERR_get_error(), NULL));
			// pthread_mutex_unlock ( &sign_mutex );
			if( resp ) OCSP_RESPONSE_free ( resp );
			resp = NULL;

			goto end;
		}
		// pthread_mutex_unlock ( &sign_mutex );
	};

	if((resp = OCSP_response_create(OCSP_RESPONSE_STATUS_SUCCESSFUL, 
								bs)) == NULL ) {
		syslog(LOG_ERR, "[Thread] Error creating response "
			"(OpenSSL Error is -> %s)", 
				ERR_error_string(ERR_get_error(),NULL));
	}

	end:

	if(nextupd) ASN1_GENERALIZEDTIME_free(nextupd);

	if( bs ) OCSP_BASICRESP_free( bs );

	return resp;
}

int ocspd_resp_send_socket(int connfd, OCSP_RESPONSE *resp, 
							OCSPD_CONFIG *conf) {
	BIO *out = NULL;
	BIO *mem = NULL;
	char * cp;
	int len;
	ASN1_GENERALIZEDTIME *date = NULL;
        ASN1_GENERALIZEDTIME *expire = NULL;

	char http_resp[] =
		"HTTP/1.0 200 OK\r\n"
		"Content-type: application/ocsp-response\r\n"
 		"Content-Transfer-Encoding: Binary\r\n"
		"Content-Length: %d\r\n";

	/*
		"Last-Modified: %d\r\n"
		"Expires: %d\r\n"
	*/

	if (!connfd) {
		syslog(LOG_ERR, "[%s::%d] ERROR, socket fd is 0!",
						__FILE__, __LINE__ );
		return 0;
	}

	if((out = BIO_new_socket(connfd, 0)) == NULL ) {
		syslog(LOG_ERR, "[%s:%d] ERROR, socket conversion to BIO",
						__FILE__, __LINE__ );
		return(0);
	}

	if((mem = BIO_new(BIO_s_mem())) == NULL ) {
		syslog(LOG_ERR, "ERROR::memory BIO creation");
		BIO_free(out);
		return(0);
	}

	date = ASN1_GENERALIZEDTIME_new();
	expire = ASN1_GENERALIZEDTIME_new();

	if( conf->set_nextUpdate ) {
		X509_gmtime_adj(expire, 
			(conf->nmin * 60) + (conf->ndays * 60 * 60 * 24) );
	} else {
		X509_gmtime_adj(expire, 0);
	}

	BIO_printf(mem, http_resp, i2d_OCSP_RESPONSE(resp, NULL));
	BIO_printf(mem, "Date: ");
	ASN1_GENERALIZEDTIME_print(mem, date);
	BIO_printf(mem, "\r\nExpires: ");
	ASN1_GENERALIZEDTIME_print(mem, expire);
	BIO_printf(mem, "\r\n\r\n");

	i2d_OCSP_RESPONSE_bio(mem, resp);
	len = BIO_get_mem_data(mem, &cp);

	if (len < 0)
		syslog(LOG_ERR, "ERROR::response length %d <= 0", len);
	else
		BIO_write(out, cp, len);

	BIO_flush(out);
	BIO_free(mem);
	BIO_free(out);

	if( date ) ASN1_GENERALIZEDTIME_free( date );
	if( expire ) ASN1_GENERALIZEDTIME_free( expire );

	return 1;
}

